# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Iterable, Optional, Tuple

from vllm import SamplingParams
from vllm.lora.request import LoRARequest
from vllm.sequence import Logprob, Sequence, SequenceGroup
from vllm.config import SchedulerConfig, CacheConfig

from llumnix.backends.vllm.scheduler import SchedulerLlumnix

def initialize_scheduler(*,
                         max_num_seqs=1000,
                         max_token_budget=1000,
                         max_model_len=1000,
                         lora_config=None) -> SchedulerLlumnix:
    block_size = 4
    scheduler_config = SchedulerConfig(max_token_budget, max_num_seqs,
                                       max_model_len)
    cache_config = CacheConfig(block_size, 1.0, 1, "auto")
    cache_config.num_cpu_blocks = 8
    cache_config.num_gpu_blocks = 8
    scheduler = SchedulerLlumnix(scheduler_config, cache_config, lora_config)
    return scheduler

def create_dummy_prompt(
    request_id: str,
    prompt_length: int,
    block_size: Optional[int] = None,
    lora_request: Optional[LoRARequest] = None,
    use_beam_search: bool = False,
    best_of: int = 1,
) -> Tuple[Sequence, SequenceGroup]:
    if not block_size:
        block_size = prompt_length

    # Create dummy prompt sequence with tokens 0...block_size-1
    # and prompt "0 ... block_size".
    prompt_tokens = list(range(prompt_length))
    prompt_str = " ".join([str(t) for t in prompt_tokens])
    prompt = Sequence(int(request_id), prompt_str, prompt_tokens, block_size)
    seq_group = SequenceGroup(
        request_id, [prompt],
        SamplingParams(use_beam_search=use_beam_search, best_of=best_of),
        time.time(), lora_request)

    return prompt, seq_group


def create_seq_group(
        seq_prompt_len: int = 1024,
        seq_output_lens: Iterable[int] = (128, ),
        request_id: str = '0',
        seq_id_start: int = 0,
        sampling_params: Optional[SamplingParams] = None) -> SequenceGroup:

    assert len(seq_output_lens) > 0

    if sampling_params is None:
        sampling_params = SamplingParams()

    prompt_token_ids = [0] * seq_prompt_len

    seqs = []
    for seq_id_offset, output_len in enumerate(seq_output_lens):
        seq = Sequence(
            seq_id=seq_id_start + seq_id_offset,
            prompt="",
            prompt_token_ids=prompt_token_ids,
            block_size=16,
        )

        for i in range(output_len):
            seq.append_token_id(
                token_id=i,
                logprobs={i: Logprob(0.0)},
            )
        seqs.append(seq)

    seq_group = SequenceGroup(
        request_id=request_id,
        seqs=seqs,
        sampling_params=sampling_params,
        arrival_time=time.time(),
    )

    return seq_group


def round_up_to_next_block(seq_len: int, block_size: int) -> int:
    return (seq_len + block_size - 1) // block_size



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from typing import Dict, List, Optional, Tuple
import torch

from vllm.config import ModelConfig, ParallelConfig
from vllm.engine.arg_utils import AsyncEngineArgs, EngineArgs
from vllm.model_executor.sampling_metadata import SamplingMetadata
from vllm.sampling_params import SamplingType
from vllm.model_executor.layers.sampler import SampleResultType, _multinomial, _greedy_sample, _random_sample,\
                                               _modify_greedy_probs_inplace, _beam_search_sample

from llumnix.logger import init_logger
from llumnix.arg_utils import EngineManagerArgs

logger = init_logger(__name__)


def detect_unsupported_feature(engine_args: EngineArgs) -> None:
    unsupported_feature = None
    if engine_args.enable_lora:
        unsupported_feature = "multi-lora serving"
    elif engine_args.enable_prefix_caching:
        unsupported_feature = "automatic prefix caching"
    elif engine_args.enable_chunked_prefill:
        unsupported_feature = "chunked prefill"
    elif engine_args.use_v2_block_manager or engine_args.speculative_model:
        unsupported_feature = "speculative decoding"
    if unsupported_feature:
        raise ValueError(f'Unsupported feature: Llumnix does not support "{unsupported_feature}" currently.')

def check_engine_args(engine_args: AsyncEngineArgs, engine_manager_args: EngineManagerArgs) -> None:
    assert engine_args.engine_use_ray and engine_args.worker_use_ray, \
            ("In Llumnix, engine and worker must be ray actor.")
    migration_config = engine_manager_args.create_migration_config()
    engine_config = engine_args.create_engine_config()
    parallel_config = engine_config.parallel_config
    if parallel_config.world_size > 1 and migration_config.migration_backend == 'nccl':
        # TODO(s5u13b): fix logger
        print("Llumnix does not support TP or PP enabled model when the migration backend is nccl, change migration backend to gloo.")
        engine_manager_args.migration_backend = 'gloo'
    detect_unsupported_feature(engine_args)

def _get_dtype_size(dtype: torch.dtype) -> int:
    return torch.tensor([], dtype=dtype).element_size()

def get_cache_block_size(
        block_size: int,
        model_config: ModelConfig,
        parallel_config: ParallelConfig,
    ) -> int:
    head_size = model_config.get_head_size()
    num_heads = model_config.get_num_kv_heads(parallel_config)
    num_layers = model_config.get_num_layers(parallel_config)

    key_cache_block = block_size * num_heads * head_size
    value_cache_block = key_cache_block
    total = num_layers * (key_cache_block + value_cache_block)
    dtype_size = _get_dtype_size(model_config.dtype)
    return dtype_size * total

# overwrite vllm function to adapt llumnix
def _sample_with_torch(
    probs: torch.Tensor,
    logprobs: torch.Tensor,
    sampling_metadata: SamplingMetadata,
    include_gpu_probs_tensor: bool,
    modify_greedy_probs: bool,
) -> Tuple[SampleResultType, Optional[torch.Tensor]]:
    categorized_seq_group_ids: Dict[SamplingType,
                                    List[int]] = {t: []
                                                  for t in SamplingType}
    categorized_sample_indices = sampling_metadata.categorized_sample_indices
    for i, seq_group in enumerate(sampling_metadata.seq_groups):
        sampling_params = seq_group.sampling_params
        sampling_type = sampling_params.sampling_type
        categorized_seq_group_ids[sampling_type].append(i)

    sample_results_dict: Dict[int, Tuple[List[int], List[int]]] = {}
    sample_metadata = {}
    multinomial_samples = {}

    # Create output tensor for sampled token ids.
    if include_gpu_probs_tensor:
        sampled_token_ids_tensor = torch.empty(logprobs.shape[0],
                                               1,
                                               dtype=torch.long,
                                               device=logprobs.device)
    else:
        sampled_token_ids_tensor = None

    # Counterintiutively, having two loops here is actually faster.
    # The first loop can run without waiting on GPU<->CPU sync.
    for sampling_type in SamplingType:
        sample_indices = categorized_sample_indices[sampling_type][:, 0]
        num_tokens = len(sample_indices)
        if num_tokens == 0:
            continue

        seq_group_id = categorized_seq_group_ids[sampling_type]
        seq_groups = [sampling_metadata.seq_groups[i] for i in seq_group_id]
        sample_metadata[sampling_type] = (seq_group_id, seq_groups)
        long_sample_indices = sample_indices.long()
        if sampling_type == SamplingType.GREEDY:
            greedy_samples = torch.argmax(logprobs[long_sample_indices],
                                          dim=-1)

            if include_gpu_probs_tensor:
                # Store sampled tokens in output tensor.
                sampled_token_ids_tensor[
                    long_sample_indices] = greedy_samples.unsqueeze(-1)

            if modify_greedy_probs:
                # If required, modify the probabilities such that sampling from
                # the modified distribution would always sample the argmax
                # token id.
                _modify_greedy_probs_inplace(logprobs, probs,
                                             long_sample_indices,
                                             greedy_samples)

        elif sampling_type in (SamplingType.RANDOM, SamplingType.RANDOM_SEED):
            max_best_of_in_batch = 1
            for seq_group in seq_groups:
                if seq_group.is_prompt:
                    sampling_params = seq_group.sampling_params
                    max_best_of_in_batch = max(max_best_of_in_batch,
                                               sampling_params.best_of)
            seeded_args = {} if sampling_type == SamplingType.RANDOM else {
                "seq_groups": seq_groups,
            }

            multinomial_samples[sampling_type] = _multinomial(
                probs[long_sample_indices], max_best_of_in_batch,
                **seeded_args)

            if include_gpu_probs_tensor:
                # Store sampled tokens in output tensor.
                sampled_token_ids_tensor[
                    long_sample_indices] = multinomial_samples[sampling_type]

        elif sampling_type == SamplingType.BEAM:
            beam_search_logprobs = logprobs[sample_indices]
        else:
            raise ValueError(f"Unsupported sampling type: {sampling_type}")

    # GPU<->CPU sync happens in the loop below.
    torch.cuda.current_stream().synchronize()
    # This also converts the sample output to Python objects.
    for sampling_type in SamplingType:
        if sampling_type not in sample_metadata:
            continue
        (seq_group_id, seq_groups) = sample_metadata[sampling_type]
        if sampling_type == SamplingType.GREEDY:
            sample_results = _greedy_sample(seq_groups, greedy_samples)
        elif sampling_type in (SamplingType.RANDOM, SamplingType.RANDOM_SEED):
            sample_results = _random_sample(seq_groups,
                                            multinomial_samples[sampling_type])
        elif sampling_type == SamplingType.BEAM:
            sample_results = _beam_search_sample(seq_groups,
                                                 beam_search_logprobs)
        sample_results_dict.update(zip(seq_group_id, sample_results))

    sample_results = [
        sample_results_dict.get(i, ([], []))
        for i in range(len(sampling_metadata.seq_groups))
    ]
    return sample_results, sampled_token_ids_tensor

def scheduler_lock(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        with self.scheduler_lock:
            return func(self, *args, **kwargs)
    return wrapper



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from typing import List

from vllm.utils import Counter
from vllm.engine.arg_utils import EngineArgs

from llumnix.logger import init_logger
from llumnix.config import MigrationConfig
from llumnix.backends.vllm.scheduler import SchedulerLlumnix
from llumnix.backends.vllm.llm_engine import LLMEngineLlumnix, BackendVLLM
from llumnix.backends.profiling import ProfilingDatabase, LatencyMemData, ProfilingResult, SimParallelConfig


logger = init_logger(__name__)

class BackendSimVLLM(BackendVLLM):
    # pylint: disable=super-init-not-called
    def __init__(
        self,
        instance_id: int,
        migration_config: MigrationConfig,
        profiling_result_file_path: str,
        gpu_type: str,
        engine_args: EngineArgs,
    ) -> None:
        # load database
        profiling_database = ProfilingDatabase(profiling_result_file_path)
        engine_config = engine_args.create_engine_config()
        model_config = engine_config.model_config
        parallel_config = engine_config.parallel_config
        model_name = model_config.model
        # get model_name from model path
        if model_name.endswith('/'):
            model_name = model_name[:-1]
        model_name = os.path.basename(model_name)
        # get latency mem
        profiling_result: ProfilingResult = profiling_database.get(model_name)
        sim_parallel_config = SimParallelConfig(gpu_type, parallel_config.tensor_parallel_size,
                                                parallel_config.pipeline_parallel_size)
        assert sim_parallel_config in profiling_result.para_dict.keys(), "sim parallel config not in database"
        latency_mem: LatencyMemData = profiling_result.para_dict[sim_parallel_config]

        self.engine: LLMEngineLlumnix = LLMEngineLlumnix.from_engine_args(migration_config=migration_config,
                                                                          latency_mem=latency_mem, engine_args=engine_args)
        self.engine.scheduler = SchedulerLlumnix(self.engine.scheduler_config, self.engine.cache_config, self.engine.lora_config)
        self.engine.output_processor.scheduler = self.engine.scheduler
        # multi-instance args
        self.migration_config = migration_config
        self.instance_id = instance_id
        self.step_counter = Counter()
        self.scaling_down = False
        self.request_server_info = {}

    def send_blocks(self, dst_ray_actor: "ray.actor.ActorHandle", src_blocks: List[int], dst_blocks: List[int]) -> None:
        self.engine.model_executor.send_blocks(len(src_blocks))



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import pytest
import numpy as np

from llumnix.instance_info import InstanceLoadCalculator, InstanceInfo
from llumnix.global_scheduler.migration_scheduler import MigrationScheduler


MIGRATE_OUT_LOAD_THRESHOLD = 3.0


def init_migration_scheduler(policy='balanced'):
    instance_load_calculator = InstanceLoadCalculator('remaining_steps', True)
    migration_scheduler = MigrationScheduler(policy, MIGRATE_OUT_LOAD_THRESHOLD, instance_load_calculator)
    return migration_scheduler

@pytest.fixture
def migration_scheduler():
    migration_scheduler = init_migration_scheduler()
    yield migration_scheduler

def test_add_instance_and_remove_instance(migration_scheduler):
    migration_scheduler.add_instance('instance_1')
    assert migration_scheduler.num_instances == 1
    migration_scheduler.add_instance('instance_2')
    assert migration_scheduler.num_instances == 2
    migration_scheduler.remove_instance('instance_1')
    assert migration_scheduler.num_instances == 1
    migration_scheduler.remove_instance('instance_2')
    assert migration_scheduler.num_instances == 0

@pytest.mark.parametrize("policy", ['balanced', 'defrag_constrained', 'defrag_relaxed'])
def test_pair_migration(policy):
    migration_scheduler = init_migration_scheduler(policy)
    num_tests = 1000
    for _ in range(num_tests):
        instance_info_dict = {}
        for instance_id in ['instance_1', 'instance_2', 'instance_3', 'instance_4']:
            instance_info = InstanceInfo()
            instance_info.instance_id = instance_id
            instance_info.instance_load_migrate = MIGRATE_OUT_LOAD_THRESHOLD + random.uniform(-1, 1)
            instance_info.num_killed_requests = random.randint(0, 1)
            instance_info.num_blocks_last_running_request = random.randint(0, 1) * np.inf
            instance_info_dict[instance_id] = instance_info
        migration_scheduler.instance_info = instance_info_dict
        migrate_instance_pairs = migration_scheduler.pair_migration()
        for migrate_out_instance, migrate_in_instance in migrate_instance_pairs:
            assert migrate_out_instance != migrate_in_instance
            if policy != 'defrag_relaxed':
                assert instance_info_dict[migrate_out_instance].num_killed_requests > 0 \
                    or instance_info_dict[migrate_out_instance].instance_load_migrate > MIGRATE_OUT_LOAD_THRESHOLD
            assert instance_info_dict[migrate_in_instance].num_killed_requests == 0 \
                and instance_info_dict[migrate_in_instance].instance_load_migrate < MIGRATE_OUT_LOAD_THRESHOLD
            if policy == 'balanced':
                assert instance_info_dict[migrate_out_instance].num_blocks_last_running_request == 0
            if instance_info_dict[migrate_out_instance].num_killed_requests == 0:
                assert instance_info_dict[migrate_out_instance].instance_load_migrate > instance_info_dict[migrate_in_instance].instance_load_migrate



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Dict
import numpy as np

from llumnix.logger import init_logger

logger = init_logger(__name__)


class InstanceInfo:
    def __init__(self,
                 num_total_gpu_blocks: int = 0,
                 num_watermark_blocks: int= 0,
                 num_used_gpu_blocks: int = 0,
                 num_free_gpu_blocks: int = 0,
                 gpu_cache_usage: float = 0.0,
                 num_running_requests: int = 0,
                 num_waiting_requests: int = 0,
                 num_killed_requests: int = 0,
                 num_blocks_first_waiting_request: int = 0,
                 waiting_time_first_waiting_request: int = 0,
                 num_blocks_all_waiting_requests: int = 0,
                 inference_type: str = "",
                 num_batched_tokens: int = 0) -> None:
        self.num_total_gpu_blocks = num_total_gpu_blocks
        self.num_watermark_blocks = num_watermark_blocks
        self.num_used_gpu_blocks = num_used_gpu_blocks
        self.num_free_gpu_blocks = num_free_gpu_blocks
        self.num_available_gpu_blocks = self.num_free_gpu_blocks - self.num_watermark_blocks
        self.gpu_cache_usage = gpu_cache_usage
        self.num_running_requests = num_running_requests
        self.num_waiting_requests = num_waiting_requests
        self.num_killed_requests = num_killed_requests
        self.num_blocks_first_waiting_request = num_blocks_first_waiting_request
        self.waiting_time_first_waiting_request = waiting_time_first_waiting_request
        self.num_blocks_all_waiting_requests = num_blocks_all_waiting_requests
        self.num_available_gpu_blocks_waiting = self.num_available_gpu_blocks - self.num_blocks_all_waiting_requests
        # For instance load computation before migration.
        self.num_blocks_last_running_request = 0

        # For global scheduling.
        self.instance_load_migrate = -np.inf
        self.instance_load_dispatch_scale = -np.inf

        # For record statistics, assigned in scheduler.
        self.inference_type = inference_type
        self.num_batched_tokens = num_batched_tokens
        self.running_seq_lens = []
        self.num_seqs = 0
        self.max_tot_tokens = 0
        self.finished_request_ids = None

        # For record statistics, assigned in backend engine.
        self.instance_id = None
        self.step_id = None
        self.timestamp = None
        self.latency = 0.0

class InstanceLoadInfo:
    def __init__(self, instance_info: InstanceInfo) -> None:
        self.num_total_gpu_blocks = instance_info.num_total_gpu_blocks
        self.num_watermark_blocks = instance_info.num_watermark_blocks
        self.num_used_gpu_blocks = instance_info.num_used_gpu_blocks
        self.num_free_gpu_blocks = instance_info.num_free_gpu_blocks
        self.num_available_gpu_blocks = instance_info.num_available_gpu_blocks

        self.num_waiting_requests = instance_info.num_waiting_requests
        self.num_running_requests = instance_info.num_running_requests
        self.num_killed_requests = instance_info.num_killed_requests

        self.num_blocks_first_waiting_request = instance_info.num_blocks_first_waiting_request
        self.waiting_time_first_waiting_request = instance_info.waiting_time_first_waiting_request
        self.num_blocks_all_waiting_requests = instance_info.num_blocks_all_waiting_requests

        self.instance_id = instance_info.instance_id
        self.step_id = instance_info.step_id

class InstanceLoadCalculator:
    def __init__(self,
                 load_metric: str,
                 enable_defrag: bool) -> None:
        self.load_metric = load_metric
        self.enable_defrag = enable_defrag
        self.load_computation_strategies: Dict[str, LoadComputationStrategy] = {
            'migrate': MigrationLoadComputation(load_metric, enable_defrag),
            'dispatch': DispatchAndScalingLoadComputation(load_metric, enable_defrag),
            'scale': DispatchAndScalingLoadComputation(load_metric, enable_defrag),
        }

    def compute_instance_load(self,
                              instance_info: InstanceInfo,
                              action: str = 'migrate') -> float:
        instance_load_info = InstanceLoadInfo(instance_info)
        assert action in self.load_computation_strategies
        load_computation_strategy = self.load_computation_strategies[action]
        return load_computation_strategy.compute_instance_load(instance_load_info)

class LoadComputationStrategy(ABC):
    def __init__(self,
                 load_metric: str,
                 enable_defrag: bool) -> None:
        self.load_metric = load_metric
        self.enable_defrag = enable_defrag

    @abstractmethod
    def compute_instance_load(self, i: InstanceLoadInfo) -> float:
        pass

class MigrationLoadComputation(LoadComputationStrategy):
    def compute_instance_load(self, i: InstanceLoadInfo) -> float:
        instance_load = -np.inf
        if self.load_metric == 'usage_ratio':
            instance_load = (i.num_used_gpu_blocks + i.num_blocks_first_waiting_request) / i.num_total_gpu_blocks
        elif self.load_metric == 'remaining_steps':
            if not self.enable_defrag:
                num_requests = i.num_running_requests
                num_available_gpu_blocks = i.num_available_gpu_blocks
            else:
                num_requests = i.num_running_requests
                if i.num_waiting_requests != 0:
                    num_requests += 1
                    # num_requests = i.num_running_requests + i.num_waiting_requests
                num_available_gpu_blocks = i.num_available_gpu_blocks - i.num_blocks_first_waiting_request
                # num_available_gpu_blocks = i.num_available_gpu_blocks - i.num_blocks_all_waiting_requests
            if num_requests == 0:
                return -np.inf
            instance_load = (num_available_gpu_blocks / num_requests)*(-1)
        return instance_load

class DispatchAndScalingLoadComputation(LoadComputationStrategy):
    def compute_instance_load(self, i: InstanceLoadInfo) -> float:
        instance_load = -np.inf
        if self.load_metric == 'usage_ratio':
            instance_load = (i.num_used_gpu_blocks + i.num_blocks_all_waiting_requests) / i.num_total_gpu_blocks
        elif self.load_metric == 'remaining_steps':
            num_requests = i.num_running_requests + i.num_waiting_requests
            num_available_gpu_blocks = i.num_available_gpu_blocks - i.num_blocks_all_waiting_requests
            if num_requests == 0:
                return -np.inf
            instance_load = (num_available_gpu_blocks / num_requests)*(-1)
        return instance_load



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Tuple, Set
from abc import ABC, abstractmethod
import copy
import numpy as np

from llumnix.logger import init_logger
from llumnix.instance_info import InstanceInfo, InstanceLoadCalculator

logger = init_logger(__name__)


class MigrationScheduler:
    def __init__(self,
                 pair_migration_policy: str,
                 migrate_out_load_threshold: float,
                 instance_load_calculator: InstanceLoadCalculator) -> None:
        self.migrate_out_load_threshold = migrate_out_load_threshold
        self.instance_load_calculator = instance_load_calculator
        self.enable_defrag = instance_load_calculator.enable_defrag
        if not self.enable_defrag:
            self.pair_migration_policy \
                = PairMigrationPolicyFactory.get_policy("balanced",
                                                        migrate_out_load_threshold=migrate_out_load_threshold,
                                                        instance_load_calculator=instance_load_calculator)
        else:
            self.pair_migration_policy \
                = PairMigrationPolicyFactory.get_policy(pair_migration_policy,
                                                        migrate_out_load_threshold=migrate_out_load_threshold,
                                                        instance_load_calculator=instance_load_calculator)

        self.num_instances = 0
        self.instance_id_set: Set[str] = set()
        # instance info args
        self.instance_info: Dict[str, InstanceInfo] = None
        self.sorted_instance_infos: List[InstanceInfo] = None

    def pair_migration(self) -> List[Tuple[str, str]]:
        self._sort_instance_infos(descending=False)
        return self.pair_migration_policy.pair_migration(self.sorted_instance_infos)

    def update_instance_infos(self,
                              instance_info: Dict[str, InstanceInfo]) -> None:
        self.instance_info = instance_info

    def add_instance(self, instance_id: str) -> None:
        self.instance_id_set.add(instance_id)
        self.num_instances = len(self.instance_id_set)

    def remove_instance(self, instance_id: str) -> None:
        self.instance_id_set.remove(instance_id)
        self.num_instances = len(self.instance_id_set)

    def _sort_instance_infos(self,
                             descending: bool = True) -> None:
        instance_infos: List[InstanceInfo] = list(self.instance_info.values())
        key_attr = 'instance_load_migrate'
        self.sorted_instance_infos = sorted(
            instance_infos,
            key=lambda instance_info: getattr(instance_info, key_attr),
            reverse=descending
        )

class PairMigrationPolicy(ABC):
    def __init__(self,
                 migrate_out_load_threshold: float,
                 instance_load_calculator: InstanceLoadCalculator) -> None:
        self.migrate_out_load_threshold = migrate_out_load_threshold
        self.instance_load_calculator = instance_load_calculator

    @abstractmethod
    def pair_migration(self,
                      sorted_instance_infos: List[InstanceInfo]
                      ) -> List[Tuple[str, str]]:
        raise NotImplementedError

class Balanced(PairMigrationPolicy):
    def pair_migration(self,
                       sorted_instance_infos: List[InstanceInfo]
                       ) -> List[Tuple[str, str]]:
        # migrate in instances
        migrate_in_instance_infos = [i for i in sorted_instance_infos
                               if i.num_killed_requests == 0 and i.instance_load_migrate < self.migrate_out_load_threshold]
        # migrate out instances
        migrate_out_instance_infos = [i for i in reversed(sorted_instance_infos)
                                if i.num_killed_requests > 0 or i.instance_load_migrate > self.migrate_out_load_threshold]
        migrate_instance_pairs = []
        for i in range(min(len(migrate_in_instance_infos), len(migrate_out_instance_infos))):
            load_diff_before_mig = migrate_out_instance_infos[i].instance_load_migrate - migrate_in_instance_infos[i].instance_load_migrate
            left_load_after_mig = self._compute_instance_load_after_migrate(migrate_in_instance_infos[i], is_migrate_in=True)
            right_load_after_mig = self._compute_instance_load_after_migrate(migrate_out_instance_infos[i], is_migrate_in=False)
            # Add some constrains to reduce unnecessary migrations
            if left_load_after_mig > self.migrate_out_load_threshold:
                continue
            load_diff_after_mig = right_load_after_mig - left_load_after_mig
            if (0 < load_diff_after_mig < load_diff_before_mig) or (migrate_in_instance_infos[i].instance_load_migrate == -np.inf):
                migrate_instance_pairs.append((migrate_out_instance_infos[i].instance_id, migrate_in_instance_infos[i].instance_id))
        return migrate_instance_pairs

    def _compute_instance_load_after_migrate(self, instance_info: InstanceInfo, is_migrate_in: bool) -> float:
        instance_info_after_migrate = copy.deepcopy(instance_info)
        num_blocks_last_running_request = instance_info_after_migrate.num_blocks_last_running_request
        if is_migrate_in:
            instance_info_after_migrate.num_running_requests += 1
            instance_info_after_migrate.num_free_gpu_blocks -= num_blocks_last_running_request
        else:
            instance_info_after_migrate.num_running_requests -= 1
            instance_info_after_migrate.num_free_gpu_blocks += num_blocks_last_running_request
        return self.instance_load_calculator.compute_instance_load(instance_info_after_migrate, action='migrate')

class DefragConstrained(PairMigrationPolicy):
    def pair_migration(self,
                       sorted_instance_infos: List[InstanceInfo]
                       ) -> List[Tuple[str, str]]:
        # migrate in instances
        migrate_in_instance_infos = [i for i in sorted_instance_infos
                               if i.num_killed_requests == 0 and i.instance_load_migrate < self.migrate_out_load_threshold]
        # migrate out instances
        migrate_out_instance_infos = [i for i in reversed(sorted_instance_infos)
                                if i.num_killed_requests > 0 or i.instance_load_migrate > self.migrate_out_load_threshold]
        migrate_instance_pairs = []
        for i in range(min(len(migrate_in_instance_infos), len(migrate_out_instance_infos))):
            # without any constrain in order to make defragmentation migrate happens as soon as possible
            migrate_instance_pairs.append((migrate_out_instance_infos[i].instance_id, migrate_in_instance_infos[i].instance_id))
        return migrate_instance_pairs

class DefragRelaxed(PairMigrationPolicy):
    def pair_migration(self,
                       sorted_instance_infos: List[InstanceInfo]
                       ) -> List[Tuple[str, str]]:
        # migrate in instances
        migrate_in_instance_infos = [i for i in sorted_instance_infos
                               if i.num_killed_requests == 0 and i.instance_load_migrate < self.migrate_out_load_threshold]
        # migrate out instances
        migrate_out_instance_infos = list(reversed(sorted_instance_infos))
        migrate_instance_pairs = []
        for i in range(min(len(migrate_in_instance_infos), len(migrate_out_instance_infos))):
            if migrate_out_instance_infos[i].num_killed_requests != 0 \
                or migrate_out_instance_infos[i].instance_load_migrate > migrate_in_instance_infos[i].instance_load_migrate:
                migrate_instance_pairs.append((migrate_out_instance_infos[i].instance_id, migrate_in_instance_infos[i].instance_id))
        return migrate_instance_pairs

class PairMigrationPolicyFactory:
    _POLICY_REGISTRY = {
        'balanced': Balanced,
        'defrag_constrained': DefragConstrained,
        'defrag_relaxed': DefragRelaxed,
    }

    @classmethod
    def get_policy(cls, policy_name: str, **kwargs) -> PairMigrationPolicy:
        return cls._POLICY_REGISTRY[policy_name](**kwargs)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Tuple, Union, Iterable, Set

from llumnix.logger import init_logger
from llumnix.config import GlobalSchedulerConfig
from llumnix.instance_info import InstanceLoadCalculator, InstanceInfo
from llumnix.global_scheduler.dispatch_scheduler import DispatchScheduler
from llumnix.global_scheduler.migration_scheduler import MigrationScheduler
from llumnix.global_scheduler.scaling_scheduler import ScalingScheduler

logger = init_logger(__name__)


class GlobalScheduler:
    def __init__(self,
                 global_scheduler_config: GlobalSchedulerConfig) -> None:
        self.global_scheduler_config = global_scheduler_config
        # instance load and instance info args
        self.load_metric = global_scheduler_config.load_metric
        self.enable_defrag = global_scheduler_config.enable_defrag
        self.instance_load_calculator = InstanceLoadCalculator(load_metric=self.load_metric,
                                                               enable_defrag=self.enable_defrag)
        # dispatch args
        self.dispatch_policy = global_scheduler_config.dispatch_policy
        self.dispatch_scheduler = DispatchScheduler(global_scheduler_config.dispatch_policy,
                                                    self.instance_load_calculator)
        # migrate args
        self.migration_scheduler = MigrationScheduler(global_scheduler_config.pair_migration_policy,
                                                      global_scheduler_config.migrate_out_load_threshold,
                                                      self.instance_load_calculator)
        # auto-scaling args
        self.scaling_scheduler = ScalingScheduler(global_scheduler_config.scale_up_threshold,
                                                  global_scheduler_config.scale_down_threshold,
                                                  global_scheduler_config.scaling_policy,
                                                  self.instance_load_calculator)

        self.num_instances = 0
        self.instance_id_set: Set[str] = set()
        self.instance_info: Dict[str, InstanceInfo] = {}

    def update_instance_infos(self, instance_infos: List[InstanceInfo]) -> None:
        for instance_info in instance_infos:
            if instance_info.instance_id in self.instance_id_set:
                # Llumnix have different instance load compuatation methods for dispatch/migrate/scale.
                instance_info.instance_load_dispatch_scale = self.instance_load_calculator.compute_instance_load(instance_info, action='dispatch')
                instance_info.instance_load_migrate = self.instance_load_calculator.compute_instance_load(instance_info, action='migrate')
                self.instance_info[instance_info.instance_id] = instance_info

    def dispatch(self) -> str:
        self.dispatch_scheduler.update_instance_infos(self.instance_info)
        instance_id = self.dispatch_scheduler.dispatch()
        return instance_id

    def pair_migration(self) -> List[Tuple[str, str]]:
        self.migration_scheduler.update_instance_infos(self.instance_info)
        migrate_instance_pairs = self.migration_scheduler.pair_migration()
        return migrate_instance_pairs

    def check_scale(self) -> Tuple[str, str]:
        self.scaling_scheduler.update_instance_infos(self.instance_info)
        scale_up_num, scale_down_num = self.scaling_scheduler.check_scale()
        return scale_up_num, scale_down_num

    def scale_up(self, instance_id: Union[str, Iterable[str]]) -> int:
        if isinstance(instance_id, str):
            instance_id = [instance_id,]
        instance_ids = list(instance_id)
        for ins_id in instance_ids:
            if ins_id not in self.instance_id_set:
                logger.info("scale up instance: {}".format(ins_id))
                new_intance_info = self._get_empty_instance_info()
                new_intance_info.instance_id = ins_id
                self.instance_info[ins_id] = new_intance_info
                self._add_instance(ins_id)
        logger.info("self.num_instances: {}, self.instances: {}".format(self.num_instances, self.instance_id_set))
        return self.num_instances

    def scale_down(self, instance_id: Union[str, Iterable[str]]) -> int:
        if isinstance(instance_id, str):
            instance_id = [instance_id,]
        instance_ids = list(instance_id)
        for ins_id in instance_ids:
            if ins_id in self.instance_id_set:
                logger.info("scale down instance: {}".format(ins_id))
                del self.instance_info[ins_id]
                self._remove_instance(ins_id)
        logger.info("self.num_instances: {}, self.instances: {}".format(self.num_instances, self.instance_id_set))
        return self.num_instances

    def _add_instance(self, instance_id: str) -> None:
        self.instance_id_set.add(instance_id)
        self.num_instances = len(self.instance_id_set)
        for scheduler in (self.dispatch_scheduler, self.migration_scheduler, self.scaling_scheduler):
            scheduler.update_instance_infos(self.instance_info)
            scheduler.add_instance(instance_id)

    def _remove_instance(self, instance_id: str) -> None:
        self.instance_id_set.remove(instance_id)
        self.num_instances = len(self.instance_id_set)
        for scheduler in (self.dispatch_scheduler, self.migration_scheduler, self.scaling_scheduler):
            scheduler.update_instance_infos(self.instance_info)
            scheduler.remove_instance(instance_id)

    def _get_empty_instance_info(self) -> InstanceInfo:
        return self.scaling_scheduler.get_empty_instance_info()



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from dataclasses import dataclass
import argparse
from typing import Tuple

from llumnix.config import GlobalSchedulerConfig, MigrationConfig


@dataclass
class EngineManagerArgs:
    launch_ray_cluster: bool = True
    disable_init_instance_by_manager: bool = False
    initial_instances: int = 1
    disable_fixed_node_init_instance: bool = False

    load_metric: str = 'remaining_steps'
    polling_interval: float = 0.05

    dispatch_policy: str = 'load'

    enable_migration: bool = True
    enable_defrag: bool = True
    pair_migration_frequency: int = 1
    pair_migration_policy: str = 'defrag_constrained'
    migrate_out_threshold: float = 3.0
    request_migration_policy: str = 'SJF'

    enable_scaling: bool = False
    min_instances: int = 1
    max_instances: int = 1
    scaling_interval: int = 10
    scaling_policy: str = 'avg_load'
    scale_up_threshold: float = 10
    scale_down_threshold: float = 60

    log_filename: str = "server.log"
    disable_log_requests_manager: bool = False
    log_instance_info: bool = True
    profiling_result_file_path: str = ""

    gpu_type: str = "a10"
    migration_backend_init_timeout: float = 10.0
    migration_backend: str = "rpc"
    migration_cache_blocks: int = 512
    migration_num_layers: int = 1
    last_stage_max_blocks: int = 16
    max_stages: int = 3

    def create_global_scheduler_configs(
        self,
    ) -> Tuple[GlobalSchedulerConfig]:
        global_scheduler_config = GlobalSchedulerConfig(self.initial_instances,
                                                        self.load_metric,
                                                        self.dispatch_policy,
                                                        self.pair_migration_policy,
                                                        self.migrate_out_threshold,
                                                        self.enable_defrag,
                                                        self.scaling_policy,
                                                        self.scale_up_threshold,
                                                        self.scale_down_threshold)
        return global_scheduler_config

    def create_migration_config(self) -> MigrationConfig:
        migration_config = MigrationConfig(self.request_migration_policy,
                                           self.migration_backend,
                                           self.migration_cache_blocks,
                                           self.migration_num_layers,
                                           self.last_stage_max_blocks,
                                           self.max_stages,
                                           self.migration_backend_init_timeout)
        return migration_config

    @classmethod
    def from_cli_args(cls, args: argparse.Namespace) -> 'EngineManagerArgs':
        # Get the list of attributes of this dataclass.
        attrs = [attr.name for attr in dataclasses.fields(cls)]
        # Set the attributes from the parsed arguments.
        engine_manager_args = cls(**{attr: getattr(args, attr) for attr in attrs})
        cls._check_args(engine_manager_args)
        return engine_manager_args

    @classmethod
    def _check_args(cls, args):
        assert args.migration_backend != 'gloo' or (args.migration_backend == 'gloo' \
            and not args.disable_init_instance_by_manager and not args.disable_fixed_node_init_instance), \
            ("When using gloo as migration backend, "
             "do not set --disable-init-instance-by-manager and --disable-fixed-node-init-instance.")

    @staticmethod
    def add_cli_args(
            parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        parser.add_argument('--disable-fixed-node-init-instance',
                            action='store_true',
                            help='disable fixing the placement of instance to current node')
        parser.add_argument('--disable-init-instance-by-manager',
                            action='store_true',
                            help='disable initializing instance by manager')
        parser.add_argument('--initial-instances',
                            type=int,
                            default=EngineManagerArgs.initial_instances,
                            help='number of instances created at initialzation')

        parser.add_argument('--load-metric',
                            type=str,
                            default=EngineManagerArgs.load_metric,
                            choices=['remaining_steps', 'usage_ratio'],
                            help='instance load metric')
        parser.add_argument('--polling-interval',
                            type=float,
                            default=EngineManagerArgs.polling_interval,
                            help='time interval(s) to update instance info and pair migration')

        parser.add_argument('--dispatch-policy',
                            type=str,
                            default=EngineManagerArgs.dispatch_policy,
                            choices=['balanced', 'load', 'queue', 'flood'],
                            help='request dispatch policy')

        parser.add_argument('--enable-migration',
                            action='store_true',
                            help='enable migrate requests between instances')
        parser.add_argument('--pair-migration-frequency',
                            type=int,
                            default=EngineManagerArgs.pair_migration_frequency,
                            help='pair migration frequency')
        parser.add_argument('--pair-migration-policy',
                            type=str,
                            default=EngineManagerArgs.pair_migration_policy,
                            choices=['balanced', 'defrag_constrained', 'defrag_relaxed'],
                            help='pair migration policy')
        parser.add_argument('--migrate-out-threshold',
                            type=float,
                            default=EngineManagerArgs.migrate_out_threshold,
                            help='migrate out instance load threshold')
        parser.add_argument('--request-migration-policy',
                            type=str,
                            default=EngineManagerArgs.request_migration_policy,
                            choices=['LCFS', 'SJF', 'LJF'],
                            help='request migration policy')
        parser.add_argument('--enable-defrag',
                            type=bool,
                            default=EngineManagerArgs.enable_defrag,
                            help='enable defragmentation through migration based on virtual usage')

        parser.add_argument('--enable-scaling',
                            action='store_true',
                            help='enable auto scaling')
        parser.add_argument('--min-instances',
                            type=int,
                            default=EngineManagerArgs.min_instances,
                            help='minimum number of instances')
        parser.add_argument('--max-instances',
                            type=int,
                            default=EngineManagerArgs.max_instances,
                            help='maximum number of instances')
        parser.add_argument('--scaling-interval',
                            type=int,
                            default=EngineManagerArgs.scaling_interval,
                            help='interval time of check scaling')
        parser.add_argument('--scaling-policy',
                            type=str,
                            default=EngineManagerArgs.scaling_policy,
                            choices=['max_load', 'avg_load'],
                            help='scaling policy')
        parser.add_argument('--scale-up-threshold',
                            type=float,
                            default=EngineManagerArgs.scale_up_threshold,
                            help='scale up threshold')
        parser.add_argument('--scale-down-threshold',
                            type=float,
                            default=EngineManagerArgs.scale_down_threshold,
                            help='scale down threshold')

        parser.add_argument('--disable-log-requests-manager',
                            action='store_true',
                            help='disable logging requests in manager')
        parser.add_argument('--log-instance-info',
                            action='store_true',
                            help='enable logging instance info')
        parser.add_argument('--log-filename',
                            type=str,
                            default=EngineManagerArgs.log_filename,
                            help='log filename')
        parser.add_argument('--profiling-result-file-path',
                            type=str,
                            default=EngineManagerArgs.profiling_result_file_path,
                            help='profiling result file path')

        parser.add_argument('--gpu-type',
                            type=str,
                            default=EngineManagerArgs.gpu_type,
                            help='gpu type specified when using simulator')

        parser.add_argument('--migration-backend',
                            type=str,
                            default=EngineManagerArgs.migration_backend,
                            choices=['gloo','nccl','rpc'],
                            help='communication backend of migration')
        parser.add_argument('--migration-backend-init-timeout',
                            type=float,
                            default=EngineManagerArgs.migration_backend_init_timeout,
                            help='timeout(s) for initializing migration backend')
        parser.add_argument('--migration-cache-blocks',
                            type=int,
                            default=EngineManagerArgs.migration_cache_blocks,
                            help='number of cache blocks in migration')
        parser.add_argument('--migration-num-layers',
                            type=int,
                            default=EngineManagerArgs.migration_num_layers,
                            help='number of kv-cache layers to transfer in each round during migration')
        parser.add_argument('--last-stage-max-blocks',
                            type=int,
                            default=EngineManagerArgs.last_stage_max_blocks,
                            help='if the number pf remain blocks < last_stage_max_blocks, do last stage migration')
        parser.add_argument('--max-stages',
                            type=int,
                            default=EngineManagerArgs.max_stages,
                            help='drop migration if the number of stages > max_stages')

        return parser



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

class MigratingRequest:
    def __init__(
            self,
            request_id: int,
            backend_request: Any,
        ) -> None:
        self.request_id = request_id
        self.backend_request = backend_request
        self.stage_timestamps = []
        self.stage_num_blocks_list = []
        self.server_info = None



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union, Iterable
import time
import ray
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy
from ray.util.scheduling_strategies import NodeAffinitySchedulingStrategy

from llumnix.logger import init_logger
from llumnix.instance_info import InstanceInfo
from llumnix.backends.backend_interface import BackendInterface, BackendType
from llumnix.backends.utils import init_backend_engine, initialize_placement_group
from llumnix.llumlet.migration_coordinator import MigrationCoordinator, MigrationStatus
from llumnix.llumlet.local_migration_scheduler import LocalMigrationScheduler
from llumnix.server_info import ServerInfo
from llumnix.config import MigrationConfig

logger = init_logger(__name__)


class Llumlet:
    def __init__(self,
                 instance_id: str,
                 backend_type: BackendType,
                 migration_config: MigrationConfig,
                 *args,
                 **kwargs) -> None:
        self.instance_id = instance_id
        self.actor_name = f"instance_{instance_id}"
        self.backend_engine: BackendInterface = init_backend_engine(self.instance_id,
                                                                    backend_type,
                                                                    migration_config,
                                                                    *args,
                                                                    **kwargs)
        self.migration_coordinator = MigrationCoordinator(self.backend_engine,
                                                          migration_config.last_stage_max_blocks,
                                                          migration_config.max_stages)
        self.migration_scheduler = LocalMigrationScheduler(migration_config.request_migration_policy,
                                                           self.backend_engine)
        self.log_requests = True

    @classmethod
    def from_args(cls,
                  disable_fixed_node_init_instance: bool,
                  detached: bool,
                  node_id: str,
                  instance_id: str,
                  backend_type: BackendType,
                  world_size: int,
                  migration_config: MigrationConfig,
                  *args,
                  **kwargs):
        lifetime = "detached" if detached else None
        assert backend_type in [backend_type.VLLM, backend_type.SIM_VLLM], f'unimplemented backend {backend_type}'
        if backend_type == backend_type.VLLM:
            if disable_fixed_node_init_instance:
                # TODO(s5u13b): Support placement_group lifetime management when the migration backend is gloo.
                placement_group = initialize_placement_group(world_size, detached=detached)
                kwargs["placement_group"] = placement_group
                engine_class = ray.remote(num_cpus=1,
                                          name=f"instance_{instance_id}",
                                          namespace='llumnix',
                                          max_concurrency=4,
                                          lifetime=lifetime)(cls).options(
                                                scheduling_strategy=PlacementGroupSchedulingStrategy(
                                                    placement_group=placement_group,
                                                    placement_group_bundle_index=0,))
            else:
                kwargs["node_id"] = node_id
                engine_class = ray.remote(num_cpus=1,
                                          name=f"instance_{instance_id}",
                                          namespace='llumnix',
                                          max_concurrency=4,
                                          lifetime=lifetime)(cls).options(
                                                scheduling_strategy=NodeAffinitySchedulingStrategy(
                                                    node_id=node_id,
                                                    soft=False,))
        else: # backend_type == backend_type.SIM_VLLM:
            engine_class = ray.remote(num_cpus=1,
                                      name=f"instance_{instance_id}",
                                      namespace='llumnix',
                                      max_concurrency=4,
                                      lifetime=lifetime)(cls).options(
                                        scheduling_strategy=NodeAffinitySchedulingStrategy(
                                            node_id=node_id,
                                            soft=False,))
        llumlet = engine_class.remote(instance_id, backend_type, migration_config, *args, **kwargs)
        return llumlet

    def migrate_out(self, dst_instance_name: str) -> List[str]:
        try:
            t0 = time.time()
            migrate_in_ray_actor = ray.get_actor(dst_instance_name, namespace='llumnix')
            dst_instance_id = dst_instance_name[len("instance_"):]
            logger.info("{}->{} begin migrate out".format(self.instance_id, dst_instance_id))
            migrate_out_request = self.migration_scheduler.get_migrate_out_request()
            migrated_request_list = []
            if migrate_out_request is None:
                return migrated_request_list
            status = self.migration_coordinator.migrate_out_multistage(migrate_in_ray_actor, migrate_out_request)
            if status == MigrationStatus.FINISHED_DONE:
                self.backend_engine.free_src_request(migrate_out_request.backend_request)
                migrated_request_list.append(migrate_out_request.request_id)
                migrate_out_request.stage_timestamps.append(time.time())
                self.backend_engine.free_request_states(migrate_out_request.request_id)
                self.backend_engine.remove_migrating_out_request_last_stage(migrate_out_request.backend_request)
            else:
                ray.get(migrate_in_ray_actor.execute_migration_method.remote("free_dst_pre_alloc_cache", migrate_out_request.request_id))
            t1 = time.time()
            logger.info("{}->{} migrate done, migrate request {}, status:{}, len:{} blocks, cost:{} ms" \
                  .format(self.instance_id, dst_instance_id, migrated_request_list, status, \
                   sum(migrate_out_request.stage_num_blocks_list), (t1 - t0)*1000))
        except ray.exceptions.RayActorError:
            logger.info("[migrate_out] instance {} is dead".format(dst_instance_name[len("instance_"):]))
            raise
        return migrated_request_list

    def get_instance_info(self) -> InstanceInfo:
        return self.backend_engine.engine.instance_info

    def is_ready(self) -> bool:
        return True

    def get_all_request_ids(self) -> List[str]:
        return self.backend_engine.get_all_request_ids()

    def generate(
        self,
        request_id: str,
        server_info: ServerInfo,
        *args,
        **kwargs,
    ) -> None:
        # This should not be used for logging, as it is monotonic time.
        self.backend_engine.add_request(request_id, server_info, *args, **kwargs)

    def abort(self, request_id: Union[str, Iterable[str]]) -> None:
        if isinstance(request_id, str):
            request_id = (request_id,)
        request_ids = set(request_id)
        return self.backend_engine.abort_request(request_ids)

    def clear_migration_states(self, is_migrate_in: bool) -> None:
        logger.info("instance {} clear_migration_states, is_migrate_in: {}".format(self.instance_id, is_migrate_in))
        if is_migrate_in:
            # If migrate out instance dies during migration, migrate in instance directly free the pre-allocated cache of the migrating in request.
            logger.info("clear_migration_states: free_dst_pre_alloc_cache")
            self.backend_engine.free_dst_pre_alloc_cache()
        else:
            # If migrate in instance dies during migration, migrate out instance should add the migrating out request in last stage.
            # back to the running request queue.
            migrating_out_requests_last_stage = self.backend_engine.pop_migrating_out_requests_last_stage()
            for backend_request in migrating_out_requests_last_stage:
                logger.info("clear_migration_states: add request {} back to engine".format(backend_request.request_id))
                self.backend_engine.add_running_request(backend_request)

    def execute_migration_method(self, method, *args, **kwargs):
        executor = getattr(self.migration_coordinator, method)
        return executor(*args, **kwargs)

    def execute_engine_method(self, method, *args, **kwargs):
        executor = getattr(self.backend_engine, method)
        return executor(*args, **kwargs)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple
import ray
# pylint: disable=unused-import
from ray.util.placement_group import PlacementGroup

from llumnix.backends.backend_interface import BackendInterface, BackendType


def init_backend_engine(instance_id: str, backend_type: BackendType, *args, **kwargs) -> BackendInterface:
    if backend_type == BackendType.VLLM:
        # pylint: disable=import-outside-toplevel
        from llumnix.backends.vllm.llm_engine import BackendVLLM
        backend_engine = BackendVLLM(instance_id, *args, **kwargs)
    elif backend_type == BackendType.SIM_VLLM:
        # pylint: disable=import-outside-toplevel
        from llumnix.backends.vllm.simulator import BackendSimVLLM
        backend_engine = BackendSimVLLM(instance_id, *args, **kwargs)
    else:
        raise ValueError(f'Unsupported backend: {backend_type}')
    return backend_engine

def initialize_placement_group(
    world_size: int = 1,
    detached: bool = False
) -> Tuple[str, Optional["PlacementGroup"]]:
    """Initialize the distributed cluster probably with Ray.

    Args:
        parallel_config: The configurations for parallel execution.
        engine_use_ray: Whether to use Ray for async engine.
        ray_address: The address of the Ray cluster. If None, uses
            the default Ray cluster address.

    Returns:
        A tuple of (`distributed_init_method`, `placement_group`). The
        `distributed_init_method` is the address for initializing the
        distributed backend. `placement_group` includes the specification
        of the resources for each distributed worker.
    """
    if ray is None:
        raise ImportError(
            "Ray is not installed. Please install Ray to use distributed "
            "serving.")

    lifetime = "detached" if detached else None
    # Create placement group for worker processes
    current_placement_group = ray.util.get_current_placement_group()
    if current_placement_group:
        # We are in a placement group
        bundles = current_placement_group.bundle_specs
        # Verify that we can use the placement group.
        gpu_bundles = 0
        for bundle in bundles:
            bundle_gpus = bundle.get("GPU", 0)
            if bundle_gpus > 1:
                raise ValueError(
                    "Placement group bundle cannot have more than 1 GPU.")
            if bundle_gpus:
                gpu_bundles += 1
        if world_size > gpu_bundles:
            raise ValueError(
                "The number of required GPUs exceeds the total number of "
                "available GPUs in the placement group.")
    else:
        num_gpus_in_cluster = ray.cluster_resources().get("GPU", 0)
        if world_size > num_gpus_in_cluster:
            raise ValueError(
                "The number of required GPUs exceeds the total number of "
                "available GPUs in the cluster.")
        # Create a new placement group
        placement_group_specs = ([{"CPU": 1}] + [{"GPU": 1}] * world_size)
        current_placement_group = ray.util.placement_group(
            placement_group_specs, "STRICT_PACK", lifetime=lifetime)
        # Wait until PG is ready - this will block until all
        # requested resources are available, and will timeout
        # if they cannot be provisioned.
        ray.get(current_placement_group.ready(), timeout=1800)

    return current_placement_group



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock, patch

import ray

from llumnix.llumlet.migration_coordinator import MigrationCoordinator
from llumnix.llumlet.migrating_request import MigratingRequest
from llumnix.backends.backend_interface import BackendInterface
from llumnix.llumlet.llumlet import MigrationStatus

@ray.remote
def ray_remote_call(ret):
    return ret

def test_migrate_out_onestage():
    # Initialize Ray
    ray.init(ignore_reinit_error=True)

    # Create mock objects
    backend_engine = MagicMock(spec=BackendInterface)
    migrate_in_ray_actor = MagicMock()
    migrate_out_request = MigratingRequest(1, "test_request")

    # Create an instance of MigrationCoordinator
    coordinator = MigrationCoordinator(backend_engine, 1, 3)

    # Mock method return values and test data
    src_blocks = [1, 2, 3]
    dst_blocks = [1, 2]
    backend_engine.get_request_incremental_blocks.return_value = src_blocks
    backend_engine.should_abort_migration.return_value = False
    migrate_in_ray_actor.execute_migration_method.remote.return_value = ray_remote_call.remote(dst_blocks)

    # Test normal migration scenario
    status = coordinator.migrate_out_onestage(migrate_in_ray_actor, migrate_out_request)
    assert status == MigrationStatus.RUNNING

    # Test the last stage of migration
    src_blocks = [3]
    dst_blocks = [3]
    backend_engine.get_request_incremental_blocks.return_value = src_blocks
    backend_engine.should_abort_migration.return_value = False
    migrate_in_ray_actor.execute_migration_method.remote.return_value = ray_remote_call.remote(dst_blocks)
    status = coordinator.migrate_out_onestage(migrate_in_ray_actor, migrate_out_request)
    assert status == MigrationStatus.FINISHED_DONE

    migrate_out_request = MigratingRequest(2, "test_request")
    # Test migration aborted scenario
    src_blocks = [1, 2, 3]
    dst_blocks = []
    backend_engine.get_request_incremental_blocks.return_value = src_blocks
    backend_engine.should_abort_migration.return_value = False
    migrate_in_ray_actor.execute_migration_method.remote.return_value = ray_remote_call.remote(dst_blocks)
    status = coordinator.migrate_out_onestage(migrate_in_ray_actor, migrate_out_request)
    assert status == MigrationStatus.FINISHED_ABORTED

    migrate_out_request = MigratingRequest(3, "test_request")
    src_blocks = [1, 2, 3]
    dst_blocks = [1, 2]
    backend_engine.get_request_incremental_blocks.return_value = src_blocks
    backend_engine.should_abort_migration.return_value = True
    migrate_in_ray_actor.execute_migration_method.remote.return_value = ray_remote_call.remote(dst_blocks)
    status = coordinator.migrate_out_onestage(migrate_in_ray_actor, migrate_out_request)
    assert status == MigrationStatus.FINISHED_ABORTED
    ray.shutdown()

@patch.object(MigrationCoordinator, 'migrate_out_onestage')
def test_migrate_out_multistage(migrate_out_onestage):
    # Initialize Ray
    ray.init(ignore_reinit_error=True)

    # Create mock objects
    backend_engine = MagicMock(spec=BackendInterface)
    migrate_in_ray_actor = MagicMock()
    migrate_out_request = MigratingRequest(1, "test_request")

    # Create an instance of MigrationCoordinator
    max_stages = 3
    coordinator = MigrationCoordinator(backend_engine, 1, max_stages)
    migrate_in_ray_actor = MagicMock()
    migrate_in_ray_actor.execute_engine_method = MagicMock()
    migrate_in_ray_actor.execute_engine_method.remote = MagicMock()
    migrate_in_ray_actor.execute_engine_method.remote.return_value = ray_remote_call.remote([1])
    migrate_in_ray_actor.execute_migration_method.remote.return_value = ray_remote_call.remote([1])
    coordinator.migrate_out_onestage.side_effect = [MigrationStatus.FINISHED_DONE]
    status = coordinator.migrate_out_multistage(migrate_in_ray_actor, migrate_out_request)
    assert coordinator.migrate_out_onestage.call_count == 1
    assert status == MigrationStatus.FINISHED_DONE

    max_stages = 3
    coordinator.migrate_out_onestage.side_effect = [MigrationStatus.RUNNING,
                                                    MigrationStatus.RUNNING,
                                                    MigrationStatus.RUNNING,
                                                    MigrationStatus.RUNNING]
    status = coordinator.migrate_out_multistage(migrate_in_ray_actor, migrate_out_request)
    assert coordinator.migrate_out_onestage.call_count == max_stages + 1
    assert status == MigrationStatus.FINISHED_ABORTED

    ray.shutdown()



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from vllm.utils import random_uuid

from llumnix.config import GlobalSchedulerConfig
from llumnix.global_scheduler.global_scheduler import GlobalScheduler
from llumnix.instance_info import InstanceInfo

from tests.global_scheduler.test_llm_engine_manager import get_instance_info_migrate_in, get_instance_info_migrate_out


def init_global_scheduler():
    global_scheduler_config = GlobalSchedulerConfig(0, 'remaining_steps', 'load', 'defrag_constrained', 3.0, True, 'avg_load', 10, 60)
    global_scheduler = GlobalScheduler(global_scheduler_config)
    return global_scheduler

def init_instance_infos(initial_instances):
    instance_infos = []
    for _ in range(initial_instances):
        instance_id = random_uuid()
        instance_info = InstanceInfo()
        instance_info.instance_id = instance_id
        instance_infos.append(instance_info)
    return instance_infos

@pytest.fixture
def global_scheduler():
    global_scheduler = init_global_scheduler()
    yield global_scheduler

def test_scale_up_and_scale_down(global_scheduler):
    initial_instances = 4
    instance_infos = init_instance_infos(initial_instances)
    instance_ids = [instance_info.instance_id for instance_info in instance_infos]
    num_instances = global_scheduler.scale_up(instance_ids)
    assert num_instances == initial_instances
    instance_infos = init_instance_infos(initial_instances)
    instance_ids_1 = [instance_info.instance_id for instance_info in instance_infos]
    num_instances = global_scheduler.scale_down(instance_ids_1)
    assert num_instances == initial_instances
    num_instances = global_scheduler.scale_down(instance_ids)
    assert num_instances == 0

def test_update_instance_infos(global_scheduler):
    initial_instances = 4
    instance_infos = init_instance_infos(initial_instances)
    global_scheduler.update_instance_infos(instance_infos)
    assert len(global_scheduler.instance_info) == 0
    instance_ids = [instance_info.instance_id for instance_info in instance_infos]
    global_scheduler.scale_up(instance_ids)
    global_scheduler.update_instance_infos(instance_infos)
    assert len(global_scheduler.instance_info) == initial_instances

def test_dispatch(global_scheduler):
    initial_instances = 4
    instance_infos = init_instance_infos(initial_instances)
    instance_ids = [instance_info.instance_id for instance_info in instance_infos]
    global_scheduler.scale_up(instance_ids)
    global_scheduler.update_instance_infos(instance_infos)
    instance_id = global_scheduler.dispatch()
    assert instance_id in instance_ids

def test_pair_migration(global_scheduler):
    instance_id = random_uuid()
    instance_id_1 = random_uuid()
    instance_ids = [instance_id, instance_id_1]
    instance_info_migrate_in = get_instance_info_migrate_in(instance_id)
    instance_info_migrate_out = get_instance_info_migrate_out(instance_id_1)
    instance_infos = [instance_info_migrate_in, instance_info_migrate_out]
    global_scheduler.scale_up(instance_ids)
    global_scheduler.update_instance_infos(instance_infos)
    migrate_instace_pairs = global_scheduler.pair_migration()
    assert migrate_instace_pairs[0][0] == instance_id_1
    assert migrate_instace_pairs[0][1] == instance_id



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List
import math
import ray
import torch

from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy, NodeAffinitySchedulingStrategy
from vllm.utils import is_pin_memory_available
from vllm.worker.worker import Worker
from vllm.config import CacheConfig,  ModelConfig, ParallelConfig
from vllm.worker.cache_engine import CacheEngine

from llumnix.logger import init_logger
from llumnix.backends.vllm.utils import _sample_with_torch
from llumnix.backends.vllm.migration_backend import MigrationBackendBase, get_migration_backend
from llumnix.config import MigrationConfig

logger = init_logger(__name__)

class MigrationWorker(Worker):
    def __init__(self, *args, **kwargs) -> None:
        # replace sampler
        # pylint: disable=import-outside-toplevel
        import vllm.model_executor.layers.sampler
        vllm.model_executor.layers.sampler._sample_with_torch = _sample_with_torch

        super().__init__(*args, **kwargs)

    def load_model(self):
        torch.cuda.set_device(self.device)
        return super().load_model()

    def get_global_rank(self):
        return self.global_rank

    def reserve_memory_for_migration(self, migration_config: MigrationConfig, model_config: ModelConfig,
                                     cache_config: CacheConfig, parallel_config: ParallelConfig) -> int:
        migrate_cache_blocks_size = migration_config.migration_cache_blocks
        migrate_num_layers = migration_config.migration_num_layers
        dummy_cache_size = migrate_num_layers * migrate_cache_blocks_size * CacheEngine.get_cache_block_size(
            cache_config, model_config, parallel_config) // model_config.get_num_layers(parallel_config)

        # For nccl migration backend, reserve gpu memory for dummy cache in migration backend. For other backends,
        # CPU memory is used for the dummy cache, which is almost unlimited, so no special action is needed.
        if migration_config.migration_backend == "nccl" and parallel_config.world_size == 1:
            device = torch.device(f"cuda:{self.local_rank}")
            _, total_memory = torch.cuda.mem_get_info(device)
            migrate_ratio = math.ceil(dummy_cache_size / total_memory * 10000) / 10000
            cache_config.gpu_memory_utilization -= migrate_ratio

            if cache_config.gpu_memory_utilization <= 0:
                raise ValueError("Nccl migration backend take {:.4f} gpu memory, which is greater than gpu_memory_utilization {:.4f}. "
                                 "try to increase gpu-memory-utilization or reduce migration-cache-blocks."
                                 .format(migrate_ratio, cache_config.gpu_memory_utilization))

            logger.info("nccl migration backend take {:.4f} gpu memory, left gpu_memory_utilization {:.4f} for kv cache."
                        .format(migrate_ratio, cache_config.gpu_memory_utilization))

        return dummy_cache_size

    def init_migration(self, instance_id: str, migration_config: MigrationConfig, src_worker_handle_list,
                       placement_group=None, node_id=None) -> None:
        if placement_group:
            scheduling_strategy = PlacementGroupSchedulingStrategy(
                placement_group=placement_group,
                placement_group_capture_child_tasks=True,
            )
        else:
            scheduling_strategy = NodeAffinitySchedulingStrategy(
                node_id=node_id,
                soft=False,
            )

        pin_memory = is_pin_memory_available()
        if not pin_memory:
            # Pinning memory in WSL is not supported.
            # https://docs.nvidia.com/cuda/wsl-user-guide/index.html#known-limitations-for-linux-cuda-applications
            logger.warning("Using 'pin_memory=False' as WSL is detected. "
                           "This may slow down the performance.")

        self.instance_id = instance_id
        self.global_world_size = 0
        self.global_rank = -1
        self.migration_backend: MigrationBackendBase = get_migration_backend(migration_config=migration_config,
                                                                             cache_engine=self.cache_engine,
                                                                             worker_handle_list=src_worker_handle_list,
                                                                             scheduling_strategy=scheduling_strategy,
                                                                             is_driver_worker=self.is_driver_worker,
                                                                             gpu_cache=self.gpu_cache,
                                                                             worker_rank=self.rank,
                                                                             local_rank=self.local_rank)

    def migrate_cache(self, src_worker_handle_list, src_blocks: List[int], dst_blocks: List[int]) -> None:
        src_worker_handle = src_worker_handle_list[self.rank]
        try:
            self.migration_backend.migrate_cache(src_worker_handle, src_blocks, dst_blocks)
        except ray.exceptions.RayActorError:
            logger.info("[migrate_cache] self.rank: {}, src_worker_handle {} is dead".format(self.rank, src_worker_handle))

    def do_recv(self, *args, **kwargs):
        return self.migration_backend.do_recv(*args, **kwargs)

    def do_send(self, *args, **kwargs):
        return self.migration_backend.do_send(*args, **kwargs)

    def rebuild_migration_backend(self, instance_rank: Dict[str, int], group_name: str) -> bool:
        self.migration_backend.destory_backend()

        ret = True
        if group_name is not None:
            num_instance = len(instance_rank)
            self.global_world_size = num_instance * self.parallel_config.world_size
            self.global_rank = self.rank + instance_rank[self.instance_id] * self.parallel_config.world_size
            ret = self.migration_backend.init_backend(group_name, self.global_world_size, self.global_rank)

        return ret

    def warmup(self) -> bool:
        return self.migration_backend.warmup()

    def shutdown(self) -> None:
        torch.cuda.synchronize()
        del self.model_runner
        del self.cache_engine
        del self.gpu_cache
        torch.cuda.empty_cache()
        torch.cuda.reset_max_memory_allocated()

    def restart(self) -> None:
        self.init_model()
        self.init_cache_engine(self.cache_config)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import ray
import pytest
import numpy as np

from vllm.utils import random_uuid

from llumnix.arg_utils import EngineManagerArgs
from llumnix.llm_engine_manager import LLMEngineManager, MANAGER_ACTOR_NAME
from llumnix.instance_info import InstanceInfo


@ray.remote(num_cpus=1, max_concurrency=4)
class MockLlumlet:
    def __init__(self, instance_id):
        self.instance_id = instance_id
        self.actor_name = f"instance_{instance_id}"
        self.num_requests = 0
        self.request_id_set = set()
        self.instance_info = None
        self.num_migrate_out = 0

    def get_instance_id(self) -> str:
        return self.instance_id

    def set_instance_info(self, instance_info):
        self.instance_info = instance_info

    def get_instance_info(self):
        return self.instance_info

    def is_ready(self) -> bool:
        return True

    def get_all_request_ids(self):
        return list(self.request_id_set)

    def get_num_requests(self):
        return self.num_requests

    def generate(self, request_id, server_info, *args, **kwargs):
        self.request_id_set.add(request_id)
        self.num_requests = len(self.request_id_set)
        return self.num_requests

    def abort(self, request_id):
        if isinstance(request_id, str):
            request_id = (request_id,)
        request_ids = set(request_id)
        for request_id in request_ids:
            if request_id in self.request_id_set:
                self.request_id_set.remove(request_id)
                self.num_requests = len(self.request_id_set)
        return self.num_requests

    def migrate_out(self, dst_instance_name):
        self.num_migrate_out += 1

    def get_num_migrate_out(self):
        return self.num_migrate_out


def init_manager():
    ray.init(ignore_reinit_error=True, namespace='llumnix')
    try:
        engine_manager_args = EngineManagerArgs()
        engine_manager_args.log_instance_info = False
        engine_manager = LLMEngineManager.from_args(engine_manager_args, None)
    except ValueError:
        engine_manager = ray.get_actor(MANAGER_ACTOR_NAME, namespace='llumnix')
    ray.get(engine_manager.is_ready.remote())
    return engine_manager

def init_llumlets(initial_instances):
    instance_ids = []
    llumlets = []
    for _ in range(initial_instances):
        instance_id = random_uuid()
        instance_name = 'instance_{}'.format(instance_id)
        llumlet = MockLlumlet.options(name=instance_name,
                                      namespace='llumnix').remote(instance_id)
        instance_ids.append(instance_id)
        llumlets.append(llumlet)
    ray.get([llumlet.is_ready.remote() for llumlet in llumlets])
    return instance_ids, llumlets

@pytest.fixture
def engine_manager():
    engine_manager = init_manager()
    ray.get(engine_manager.is_ready.remote())
    yield engine_manager
    ray.kill(engine_manager)
    ray.shutdown()

@pytest.fixture
def llumlet():
    instance_id = random_uuid()
    instance_name = 'instance_{}'.format(instance_id)
    llumlet = MockLlumlet.options(name=instance_name,
                                  namespace='llumnix').remote(instance_id)
    ray.get(llumlet.is_ready.remote())
    return llumlet

def test_init_manager(engine_manager):
    assert engine_manager is not None
    engine_manager_actor_handle = ray.get_actor(MANAGER_ACTOR_NAME, namespace='llumnix')
    assert engine_manager_actor_handle is not None
    assert engine_manager == engine_manager_actor_handle

def test_init_llumlet(llumlet):
    assert llumlet is not None
    ray.get(llumlet.is_ready.remote())

# TODO(s5u13b): Add init_llumlets test.

def test_scale_up_and_down(engine_manager):
    initial_instances = 4
    instance_ids, llumlets = init_llumlets(initial_instances)
    num_instances = ray.get(engine_manager.scale_up.remote(instance_ids, llumlets))
    assert num_instances == initial_instances
    instance_ids_1, llumlets_1 = init_llumlets(initial_instances)
    num_instances = ray.get(engine_manager.scale_down.remote(instance_ids_1))
    assert num_instances == initial_instances
    num_instances = ray.get(engine_manager.scale_up.remote(instance_ids_1, llumlets_1))
    assert num_instances == initial_instances * 2
    num_instances = ray.get(engine_manager.scale_down.remote(instance_ids))
    assert num_instances == initial_instances
    num_instances = ray.get(engine_manager.scale_down.remote(instance_ids_1))
    assert num_instances == 0

def test_connect_to_instances():
    initial_instances = 4
    instance_ids, llumlets = init_llumlets(initial_instances)
    ray.get([llumlet.is_ready.remote() for llumlet in llumlets])
    engine_manager = init_manager()
    instance_ids_1, llumlets_1 = init_llumlets(initial_instances)
    num_instances = ray.get(engine_manager.scale_up.remote(instance_ids_1, llumlets_1))
    assert num_instances == initial_instances * 2
    num_instances = ray.get(engine_manager.scale_down.remote(instance_ids))
    assert num_instances == initial_instances
    ray.kill(engine_manager)
    ray.shutdown()

def test_generate_and_abort(engine_manager, llumlet):
    instance_id = ray.get(llumlet.get_instance_id.remote())
    ray.get(engine_manager.scale_up.remote(instance_id, [llumlet]))
    request_id = random_uuid()
    num_requests = ray.get(llumlet.get_num_requests.remote())
    assert num_requests == 0
    ray.get(engine_manager.generate.remote(request_id, None, None, None))
    num_requests = ray.get(llumlet.get_num_requests.remote())
    assert num_requests == 1
    ray.get(engine_manager.abort.remote(request_id))
    num_requests = ray.get(llumlet.get_num_requests.remote())
    assert num_requests == 0
    request_id_1 = random_uuid()
    request_id_2 = random_uuid()
    request_ids = [request_id_1, request_id_2]
    ray.get(engine_manager.abort.remote(request_ids))
    num_requests = ray.get(llumlet.get_num_requests.remote())
    assert num_requests == 0

def test_get_request_instance():
    _, llumlets = init_llumlets(2)
    llumlet, llumlet_1 = llumlets[0], llumlets[1]
    request_id = random_uuid()
    request_id_1 = random_uuid()
    ray.get(llumlet.generate.remote(request_id, None, None, None))
    ray.get(llumlet_1.generate.remote(request_id_1, None, None, None))
    num_requests = ray.get(llumlet.get_num_requests.remote())
    num_requests_1 = ray.get(llumlet_1.get_num_requests.remote())
    assert num_requests == 1
    assert num_requests_1 == 1
    engine_manager = init_manager()
    ray.get(engine_manager.abort.remote(request_id))
    ray.get(engine_manager.abort.remote(request_id_1))
    num_requests = ray.get(llumlet.get_num_requests.remote())
    num_requests_1 = ray.get(llumlet_1.get_num_requests.remote())
    assert num_requests == 0
    assert num_requests_1 == 0
    ray.kill(engine_manager)
    ray.shutdown()

def get_instance_info_migrate_in(instance_id):
    instance_info = InstanceInfo()
    instance_info.instance_id = instance_id
    instance_info.num_available_gpu_blocks = np.inf
    instance_info.num_running_requests = 1
    instance_info.num_blocks_first_waiting_request = 0
    return instance_info

def get_instance_info_migrate_out(instance_id):
    instance_info = InstanceInfo()
    instance_info.instance_id = instance_id
    instance_info.num_available_gpu_blocks = 0
    instance_info.num_running_requests = 1
    instance_info.num_blocks_first_waiting_request = np.inf
    return instance_info

def test_update_instance_info_loop_and_migrate(engine_manager):
    instance_ids, llumlets = init_llumlets(2)
    instance_id, instance_id_1 = instance_ids[0], instance_ids[1]
    llumlet, llumlet_1 = llumlets[0], llumlets[1]
    request_id = random_uuid()
    request_id_1 = random_uuid()
    ray.get(llumlet.generate.remote(request_id, None, None, None))
    ray.get(llumlet_1.generate.remote(request_id_1, None, None, None))
    instance_info_migrate_out = get_instance_info_migrate_out(instance_id)
    instance_info_migrate_in = get_instance_info_migrate_in(instance_id_1)
    ray.get(llumlet.set_instance_info.remote(instance_info_migrate_out))
    ray.get(llumlet_1.set_instance_info.remote(instance_info_migrate_in))
    num_migrate_out = ray.get(llumlet.get_num_migrate_out.remote())
    assert num_migrate_out == 0
    ray.get(engine_manager.scale_up.remote(instance_ids, llumlets))
    time.sleep(0.2)
    num_migrate_out = ray.get(llumlet.get_num_migrate_out.remote())
    assert num_migrate_out != 0



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Set
from abc import ABC, abstractmethod
import random

from llumnix.logger import init_logger
from llumnix.instance_info import InstanceLoadCalculator, InstanceInfo

logger = init_logger(__name__)


class DispatchScheduler:
    def __init__(self,
                 dispatch_policy: str,
                 instance_load_calculator: InstanceLoadCalculator) -> None:
        self.dispatch_policy = DispatchPolicyFactory.get_policy(dispatch_policy)
        self.instance_load_calculator = instance_load_calculator
        self.num_instances = 0
        self.instance_id_set: Set[str] = set()
        # instance info args
        self.instance_info: Dict[str, InstanceInfo] = {}
        self.sorted_instance_infos: List[InstanceInfo] = None
        # statistics
        self.num_requests = 0
        self.instance_num_requests: Dict[str, int] = {}

    def dispatch(self) -> str:
        self.num_requests += 1
        if isinstance(self.dispatch_policy, (Load, Queue)):
            self._sort_instance_infos(descending=False)
        dispatch_instance_id = self.dispatch_policy.dispatch(self.instance_num_requests,
                                                             self.sorted_instance_infos)
        self.instance_num_requests[dispatch_instance_id] += 1
        if self.num_requests % 100 == 0:
            logger.info("self.num_requests: {}".format(self.num_requests))
            for instance_id, num_requests in self.instance_num_requests.items():
                logger.info("Instance {} num_dispatched_requests: {}".format(instance_id, num_requests))
        return dispatch_instance_id

    def update_instance_infos(self,
                              instance_info: Dict[str, InstanceInfo]) -> None:
        self.instance_info = instance_info

    def add_instance(self, instance_id: str) -> None:
        self.instance_id_set.add(instance_id)
        self.num_instances = len(self.instance_id_set)
        self.instance_num_requests[instance_id] = 0

    def remove_instance(self, instance_id: str) -> None:
        self.instance_id_set.remove(instance_id)
        self.num_instances = len(self.instance_id_set)
        del self.instance_num_requests[instance_id]

    def _sort_instance_infos(self,
                            descending: bool = True) -> None:
        instance_infos: List[InstanceInfo] = list(self.instance_info.values())
        if isinstance(self.dispatch_policy, Queue):
            key_attr = 'num_waiting_requests'
        else:
            key_attr = 'instance_load_dispatch_scale'
        self.sorted_instance_infos = sorted(
            instance_infos,
            key=lambda instance_info: getattr(instance_info, key_attr),
            reverse=descending
        )

class DispatchPolicy(ABC):
    def __init__(self):
        self.instance_ptr = 0

    @abstractmethod
    def dispatch(self,
                 instance_num_requests: Dict[str, int],
                 sorted_instance_infos: List[InstanceInfo]) -> int:
        pass

# Dispatch all requests to a single instance, used only for testing
class Flood(DispatchPolicy):
    def dispatch(self,
                 instance_num_requests: Dict[str, int],
                 sorted_instance_infos: List[InstanceInfo]) -> str:
        instance_id = max(instance_num_requests, key=instance_num_requests.get)
        return instance_id

class Balanced(DispatchPolicy):
    def dispatch(self,
                 instance_num_requests: Dict[str, int],
                 sorted_instance_infos: List[InstanceInfo]) -> str:
        # dispatch request according to the number of requests dispatched to instance by manager
        instance_id = min(instance_num_requests, key=instance_num_requests.get)
        return instance_id

class Load(DispatchPolicy):
    def dispatch(self,
                 instance_num_requests: Dict[str, int],
                 sorted_instance_infos: List[InstanceInfo]) -> str:
        instance_id = sorted_instance_infos[0].instance_id
        logger.info("dispatch to {}, load: {}".format(instance_id, sorted_instance_infos[0].instance_load_dispatch_scale))
        return instance_id

class Queue(DispatchPolicy):
    def dispatch(self,
                 instance_num_requests: Dict[str, int],
                 sorted_instance_infos: List[InstanceInfo]) -> str:
        min_queue_size = sorted_instance_infos[0].num_waiting_requests
        instance_id_list = []
        for instance_info in sorted_instance_infos:
            if instance_info.num_waiting_requests == min_queue_size:
                instance_id_list.append(instance_info.instance_id)
        instance_id = random.choice(instance_id_list)
        logger.info("dispatch to {}, queue size: {}".format(instance_id, sorted_instance_infos[0].num_waiting_requests))
        return instance_id

class DispatchPolicyFactory:
    _POLICY_REGISTRY = {
        'flood': Flood,
        'balanced': Balanced,
        'load': Load,
        'queue': Queue,
    }

    @classmethod
    def get_policy(cls, policy_name: str, **kwargs) -> DispatchPolicy:
        return cls._POLICY_REGISTRY[policy_name](**kwargs)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import enum
from typing import List

import ray

from llumnix.logger import init_logger
from llumnix.llumlet.migrating_request import MigratingRequest
from llumnix.backends.backend_interface import BackendInterface

logger = init_logger(__name__)

class MigrationStatus(enum.Enum):
    """Status of Migration."""
    RUNNING = enum.auto()
    FINISHED_ABORTED = enum.auto()
    FINISHED_DONE = enum.auto()

    @staticmethod
    def is_finished(status: "MigrationStatus") -> bool:
        return status in [
            MigrationStatus.FINISHED_ABORTED,
            MigrationStatus.FINISHED_DONE
        ]

class MigrationCoordinator:
    def __init__(self,
                 backend_engine: BackendInterface,
                 last_stage_max_blocks: int,
                 max_stages: int) -> None:
        self.last_stage_max_blocks = last_stage_max_blocks
        self.max_stages = max_stages
        self.backend_engine = backend_engine

    def migrate_out_onestage(self, migrate_in_ray_actor: "ray.actor.ActorHandle",  migrate_out_request: MigratingRequest, ) -> "MigrationStatus":
        """one-stage live migration until last stage
        """
        pre_stage_num_blocks = sum(migrate_out_request.stage_num_blocks_list)
        incremental_blocks = self.backend_engine.get_request_incremental_blocks(migrate_out_request.backend_request, pre_stage_num_blocks)
        # live migration, transfer all blocks except last one(currently updating)
        migration_status = MigrationStatus.RUNNING
        is_last_stage = (len(incremental_blocks) <= self.last_stage_max_blocks)
        if not is_last_stage:
            src_blocks = incremental_blocks[:-1]
            stage_block_num = len(incremental_blocks) - 1
            dst_blocks = ray.get(migrate_in_ray_actor.execute_migration_method \
                            .remote("migrate_in_pre_alloc", migrate_out_request.request_id, stage_block_num))
        else:
            # last stage migration, stop inference, transfer all blocks
            migration_status = MigrationStatus.FINISHED_DONE
            self.backend_engine.remove_running_request(migrate_out_request.request_id)
            self.backend_engine.add_migrating_out_request_last_stage(migrate_out_request.backend_request)
            stage_block_num = len(incremental_blocks)
            src_blocks = incremental_blocks[:]
            migrate_out_request.server_info = self.backend_engine.get_request_server_info(migrate_out_request.request_id)
            dst_blocks = ray.get(migrate_in_ray_actor.execute_migration_method \
                            .remote("migrate_in_last_stage", migrate_out_request, stage_block_num))

        if len(dst_blocks) != len(src_blocks):
            # migrate-in instance failed to prev alloc
            if is_last_stage:
                self.backend_engine.add_running_request(migrate_out_request.backend_request)
                self.backend_engine.remove_migrating_out_request_last_stage(migrate_out_request.backend_request)
            migration_status = MigrationStatus.FINISHED_ABORTED
            return migration_status
        # do stage send/recv
        migrate_out_request.stage_timestamps.append(time.time())
        migrate_out_request.stage_num_blocks_list.append(stage_block_num)
        # TODO(ZeldaHuang): send_blocks in migrate_in_pre_alloc/migrate_in_last_stage
        self.backend_engine.send_blocks(migrate_in_ray_actor, src_blocks, dst_blocks)
        if not is_last_stage and self.backend_engine.should_abort_migration(migrate_out_request.backend_request, \
                                                                            migrate_out_request.stage_timestamps[-1]):
            # migrate-out request abort by scheduler during send/recv
            migration_status = MigrationStatus.FINISHED_ABORTED

        return migration_status

    def migrate_out_multistage(self, migrate_in_ray_actor: "ray.actor.ActorHandle",  migrate_out_request: MigratingRequest) -> "MigrationStatus":
        """Migrate out requests to a specified instance, return migrated request id.
        Args:
        dst_instance_name:instance actor name, used to get ray actor handle
        """
        state_count = 0
        while state_count < self.max_stages:
            state_count += 1
            status = self.migrate_out_onestage(migrate_in_ray_actor, migrate_out_request)
            if MigrationStatus.is_finished(status):
                break
        return status

    def migrate_in_last_stage(self, request: MigratingRequest, block_num: int) -> List[int]:
        """last stage migrate in, update migrate in dict
        """
        pre_alloc_blocks = self.migrate_in_pre_alloc(request.request_id, block_num)
        if len(pre_alloc_blocks) == block_num:
            # Pass the server information of the request to dst instance.
            self.backend_engine.commit_dst_request(request.backend_request, request.server_info)
        return pre_alloc_blocks

    def migrate_in_pre_alloc(self, request_id: str, block_num: int) -> List[int]:
        """prev alloc blocks to migrate in request
        """
        pre_alloc_blocks = self.backend_engine.pre_alloc(request_id ,block_num)
        if len(pre_alloc_blocks) != block_num:
            # failed to alloc, abort request
            self.free_dst_pre_alloc_cache(request_id)
        return pre_alloc_blocks

    def free_dst_pre_alloc_cache(self, request_id: str = None) -> None:
        self.backend_engine.free_dst_pre_alloc_cache(request_id)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
import os
from typing import List

ROOT_DIR = os.path.dirname(__file__)

def get_path(*filepath) -> str:
    return os.path.join(ROOT_DIR, *filepath)

def get_requirements() -> List[str]:
    """Get Python package dependencies from requirements.txt."""
    with open(get_path("requirements.txt")) as f:
        requirements = f.read().strip().split("\n")
    return requirements

def readme():
    with open('README.md', encoding='utf-8') as f:
        content = f.read()
    return content

setup(
    name='llumnix',
    version='0.0.2',
    python_requires='>=3.8.1, <3.11',
    description='Efficient and easy multi-instance LLM serving',
    long_description=readme(),
    long_description_content_type="text/markdown",
    author='Llumnix Team',
    url='https://github.com/AlibabaPAI/llumnix',
    license="Apache 2.0",
    packages=find_packages(),
    install_requires=get_requirements(),
    platforms=["all"],
    classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          "License :: OSI Approved :: Apache Software License",
          "Topic :: Scientific/Engineering :: Artificial Intelligence",
      ],
)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import uvicorn
import ray
from fastapi.responses import JSONResponse, Response
from ray.util.queue import Queue as RayQueue

from vllm.outputs import CompletionOutput, RequestOutput

import llumnix.entrypoints.vllm.api_server
import llumnix.llm_engine_manager
from llumnix.arg_utils import EngineManagerArgs


app = llumnix.entrypoints.vllm.api_server.app
engine_manager = None
request_output_queue = RayQueue()
llumnix.entrypoints.vllm.api_server.request_output_queue = request_output_queue
MANAGER_ACTOR_NAME = llumnix.llm_engine_manager.MANAGER_ACTOR_NAME


@ray.remote(num_cpus=0)
class MockLLMEngineManager:
    def __init__(self):
        self._num_generates = 0
        self._num_aborts = 0

    async def generate(self, request_id, server_info, *args, **kwargs):
        self._num_generates += 1
        completion_output = CompletionOutput(0, "", [], 0.0, None)
        request_output = RequestOutput(request_id, "", [], None, [completion_output], finished=True)
        request_output_queue.put(request_output)

    async def abort(self, request_id):
        self._num_aborts += 1

    def testing_stats(self):
        return {"num_aborted_requests": self._num_aborts}


def init_manager():
    engine_manager = MockLLMEngineManager.options(name=MANAGER_ACTOR_NAME,
                                                  namespace='llumnix').remote()
    return engine_manager

@app.get("/stats")
def stats() -> Response:
    """Get the statistics of the engine."""
    return JSONResponse(ray.get(engine_manager.testing_stats.remote()))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--port", type=int, default=8000)
    parser = EngineManagerArgs.add_cli_args(parser)
    args = parser.parse_args()

    engine_manager = init_manager()
    llumnix.entrypoints.vllm.api_server.engine_manager = engine_manager

    uvicorn.run(
        app,
        host=args.host,
        port=args.port,
        log_level="debug",
        timeout_keep_alive=llumnix.entrypoints.vllm.api_server.TIMEOUT_KEEP_ALIVE)



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch
import ray
from ray.util.queue import Queue as RayQueue
from ray.util.scheduling_strategies import NodeAffinitySchedulingStrategy

from vllm import EngineArgs, SamplingParams
from vllm.utils import random_uuid

from llumnix.backends.vllm.llm_engine import BackendVLLM
from llumnix.llumlet.llumlet import Llumlet
from llumnix.backends.utils import BackendType
from llumnix.config import MigrationConfig
from llumnix.server_info import ServerInfo

from .test_llm_engine import MockEngine
from .utils import create_dummy_prompt

TEST_PROMPTS = ["hello world, ",
                "Briefly describe the major milestones in the development of artificial intelligence from 1950 to 2020.\n",
                "Write a short story about a robot that dreams for the first time.\n",
                "Explain the cultural significance of the Mona Lisa painting, and how its perception might vary in Western versus Eastern societies.\n",
                "Swahili: 'The early bird catches the worm.'\n"]

class MockBackendVLLM(BackendVLLM):
    def __init__(self):
        self.engine = MockEngine()

class MockLlumlet(Llumlet):
    def __init__(self):
        self.instance_id = "0"
        self.backend_engine = MockBackendVLLM()

@pytest.mark.skipif(torch.cuda.device_count() < 2,
                    reason="Need at least 2 GPUs to run the test.")
def test_migration_correctness():
    ray.init(namespace="llumnix", ignore_reinit_error=True)
    engine_args = EngineArgs(model="facebook/opt-125m",worker_use_ray=True)
    id_rank_map = {"0":0,"1":1}
    migration_config = MigrationConfig("LCFS", "gloo",16,1,4,5,20)
    que = RayQueue(actor_options={
        "scheduling_strategy": NodeAffinitySchedulingStrategy(
            node_id=ray.get_runtime_context().get_node_id(),
            soft=False,)
    })
    server_info = ServerInfo("0",que)

    llumlet_0:Llumlet = Llumlet.from_args(
                            False,
                            True,
                            ray.get_runtime_context().get_node_id(),
                            "0",
                            BackendType.VLLM,
                            1,
                            migration_config,
                            engine_args,)

    llumlet_1:Llumlet = Llumlet.from_args(
                            False,
                            True,
                            ray.get_runtime_context().get_node_id(),
                            "1",
                            BackendType.VLLM,
                            1,
                            migration_config,
                            engine_args,
                     )
    while True:
        res = ray.get([llumlet_0.is_ready.remote(),llumlet_1.is_ready.remote()])
        if all(res):
            break
    ray.get([llumlet_0.execute_engine_method.remote("_run_workers","rebuild_migration_backend", id_rank_map, "llumnix"),
            llumlet_1.execute_engine_method.remote("_run_workers","rebuild_migration_backend", id_rank_map, "llumnix")])
    # empty instance migrate out
    res = ray.get(llumlet_0.migrate_out.remote("instance_1"))
    assert not res

    # running without migration
    def test_correctness(prompt):
        sampling_params = SamplingParams(top_k=1, temperature=0, ignore_eos=True, max_tokens=100)
        request_id0 = random_uuid()
        llumlet_0.generate.remote(request_id0, server_info, prompt, sampling_params)
        request_output_queue = que
        origin_output = None
        finished = False
        while not finished:
            qsize = ray.get(request_output_queue.actor.qsize.remote())
            request_outputs = ray.get(request_output_queue.actor.get_nowait_batch.remote(qsize))
            for request_output in request_outputs:
                origin_output = request_output.outputs[0]
                finished = request_output.finished

        request_id1 = random_uuid()
        llumlet_0.generate.remote(request_id1, server_info, prompt, sampling_params)
        # wait prefill done
        while True:
            if ray.get(llumlet_0.execute_engine_method.remote("get_last_running_request")):
                break
        # migrate request
        res = ray.get(llumlet_0.migrate_out.remote("instance_1"))
        assert len(res) == 1
        request_output_queue = que
        output = None
        finished = False
        while not finished:
            qsize = ray.get(request_output_queue.actor.qsize.remote())
            request_outputs = ray.get(request_output_queue.actor.get_nowait_batch.remote(qsize))
            for request_output in request_outputs:
                if request_output.request_id != request_id1:
                    continue
                output = request_output.outputs[0]
                finished = request_output.finished
        assert output.text == origin_output.text
        assert output.cumulative_logprob == origin_output.cumulative_logprob
    for prompt in TEST_PROMPTS:
        test_correctness(prompt)
    ray.shutdown()

def test_clear_migration_states():
    llumlet = MockLlumlet()
    llumlet.backend_engine.pre_alloc("0", 1)
    num_gpu_blocks = 8
    block_size = 4

    llumlet.clear_migration_states(is_migrate_in=True)
    assert len(llumlet.backend_engine.pre_alloc("0", num_gpu_blocks)) == num_gpu_blocks
    _, seq_group = create_dummy_prompt("0",7,block_size)
    llumlet.backend_engine.add_migrating_out_request_last_stage(seq_group)
    llumlet.clear_migration_states(is_migrate_in=False)
    assert llumlet.backend_engine.get_last_running_request() is not None



# Copyright (c) 2024, Alibaba Group;
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, AsyncGenerator
from contextlib import asynccontextmanager
import argparse
import time
import asyncio
import json
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse, Response, StreamingResponse
import uvicorn
import ray

from vllm.sampling_params import SamplingParams
from vllm.engine.arg_utils import AsyncEngineArgs
from vllm.engine.async_llm_engine import AsyncStream

from llumnix.arg_utils import EngineManagerArgs
from llumnix.server_info import ServerInfo
from llumnix.entrypoints.llumnix_utils import (launch_ray_cluster, connect_to_ray_cluster,
                                                is_gpu_available, init_llumnix_components)
from llumnix.logger import init_logger
from llumnix.utils import random_uuid
from llumnix.backends.vllm.utils import check_engine_args

logger = init_logger(__name__)
engine_manager = None
instances = {}
instance_num_requests: Dict[str, int] = {}
# request_output_queue could be None if initialzed in lifespan.
request_output_queue = None
server_id = None
TIMEOUT_KEEP_ALIVE = 5  # seconds.
request_streams: Dict[str, AsyncStream] = {}
log_requests = None
num_finished_requests = 0
WAIT_MANAGER_INTERVAL = 5
manager_available = True


async def _background_process_outputs():
    while True:
        qsize = await request_output_queue.actor.qsize.remote()
        request_outputs = await request_output_queue.actor.get_nowait_batch.remote(qsize)
        for request_output in request_outputs:
            request_id = request_output.request_id
            # Request could be dispatched twice when manager is dead, the first request will free the request_streams when finished.
            if request_id not in request_streams:
                continue
            request_streams[request_id].put(request_output)
            if request_output.finished:
                request_streams[request_id].finish()
                del request_streams[request_id]

# pylint: disable=unused-argument
@asynccontextmanager
async def lifespan(fastapi_app: FastAPI):
    asyncio.create_task(_background_process_outputs())
    yield

app = FastAPI(lifespan=lifespan)

async def manager_generate(prompt, sampling_params, request_id) -> AsyncStream:
    if sampling_params.n > 1 or sampling_params.use_beam_search:
        raise ValueError("Unsupported feature: multiple sequence decoding")
    results_generator = AsyncStream(request_id)
    request_streams[request_id] = results_generator
    # This request's outputs will be put to the request_output_queue of this api server no matter which instance it's running in.
    server_info = ServerInfo(server_id, request_output_queue)
    # If manager is unavailable, request will be directly added to the llumlet held by api server.
    global manager_available
    try:
        # await to catch exception
        await engine_manager.generate.remote(request_id, server_info, prompt, sampling_params)
        manager_available = True
    except ray.exceptions.RayActorError:
        # Do not re-generate the request to avoid duplicate requests.
        if manager_available:
            manager_available = False
            return results_generator
        try:
            if instance_num_requests:
                instance_id = min(instance_num_requests, key=instance_num_requests.get)
                instance_num_requests[instance_id] += 1
                await instances[instance_id].generate.remote(request_id, server_info, prompt, sampling_params)
                print("Manager is unavailable, directly pass request {} to instance {}".format(request_id, instance_id))
            else:
                print("Manager is unavailable, but there is no instance behind this api server, "
                      "sleep {}s, waiting for manager restarts".format(WAIT_MANAGER_INTERVAL))
                await asyncio.sleep(WAIT_MANAGER_INTERVAL)
                return await asyncio.create_task(manager_generate(prompt, sampling_params, request_id))
        except (ray.exceptions.RayActorError, KeyError):
            if instance_id in instances:
                print("[manager_generate] instance {} is dead".format(instance_id))
                del instances[instance_id]
                del instance_num_requests[instance_id]
            return await asyncio.create_task(manager_generate(prompt, sampling_params, request_id))
    return results_generator

async def manager_abort(request_id: str) -> None:
    try:
        print("abort request: {}.".format(request_id))
        await engine_manager.abort.remote(request_id)
    except ray.exceptions.RayActorError:
        print("Manager is unavailable")


@app.get("/health")
async def health() -> Response:
    """Health check."""
    return Response(status_code=200)


@app.post("/generate")
async def generate(request: Request) -> Response:
    """Generate completion for the request.

    The request should be a JSON object with the following fields:
    - prompt: the prompt to use for the generation.
    - stream: whether to stream the results or not.
    - other fields: the sampling parameters (See `SamplingParams` for details).
    """
    request_dict = await request.json()
    prompt = request_dict.pop("prompt")
    stream = request_dict.pop("stream", False)
    sampling_params = SamplingParams(**request_dict)
    request_id = random_uuid()

    results_generator = await manager_generate(prompt, sampling_params, request_id)

    # Streaming case
    async def stream_results() -> AsyncGenerator[bytes, None]:
        async for request_output in results_generator:
            prompt = request_output.prompt
            text_outputs = [
                prompt + output.text for output in request_output.outputs
            ]
            ret = {"text": text_outputs}
            yield (json.dumps(ret) + "\0").encode("utf-8")

    if stream:
        return StreamingResponse(stream_results())

    # Non-streaming case
    final_output = None
    async for request_output in results_generator:
        if await request.is_disconnected():
            # Abort the request if the client disconnects.
            await manager_abort(request_id)
            return Response(status_code=499)
        final_output = request_output

    assert final_output is not None
    prompt = final_output.prompt
    text_outputs = [prompt + output.text for output in final_output.outputs]
    ret = {"text": text_outputs}
    return JSONResponse(ret)


@app.post("/generate_benchmark")
async def generate_benchmark(request: Request) -> Response:
    request_dict = await request.json()
    prompt = request_dict.pop("prompt")
    _ = request_dict.pop("stream", False)
    sampling_params = SamplingParams(**request_dict)
    request_id = random_uuid()

    results_generator = await manager_generate(prompt, sampling_params, request_id)

    per_token_latency = []
    start = time.time()

    # Non-streaming case
    final_output = None
    async for request_output in results_generator:
        if await request.is_disconnected():
            # Abort the request if the client disconnects.
            await manager_abort(request_id)
            return Response(status_code=499)
        now_time = time.time()
        per_token_latency.append([now_time, int((now_time - start)*1000)])
        start = now_time
        final_output = request_output

    global num_finished_requests
    if log_requests:
        # TODO(s5u13b): Use logger.
        print(f"Finished request {request_id}.")
        num_finished_requests += 1
        print(f"num_finished_requests {num_finished_requests}.")

    generation = final_output.outputs[0].text
    num_output_tokens = len(final_output.outputs[0].token_ids)
    num_input_tokens = len(final_output.prompt_token_ids)
    expected_resp_len = request_dict['max_tokens']
    if not max(expected_resp_len, 1) == max(num_output_tokens, 1):
        "request_id={}, expected_resp_len={}, num_output_tokens={}, num_input_tokens={}".format(
            request_id, expected_resp_len, num_output_tokens, num_input_tokens)
    ret = {
        'generated_text': generation,
        'num_output_tokens_cf': num_output_tokens,
        'per_token_latency': per_token_latency,
        'error': None,
        'request_id': request_id
    }
    return JSONResponse(ret)


@app.get("/is_ready")
async def is_ready():
    ready_status = await engine_manager.is_ready.remote()
    return ready_status

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--port", type=int, default=8000)
    parser.add_argument("--ssl-keyfile", type=str, default=None)
    parser.add_argument("--ssl-certfile", type=str, default=None)
    parser.add_argument('--disable-log-requests-server',
                        action='store_true',
                        help='disable logging requests in server')
    parser.add_argument("--ray-cluster-port", type=int, default=30050)
    parser.add_argument('--launch-ray-cluster',
                        action='store_true',
                        help='if launch ray cluster in api server')

    parser = EngineManagerArgs.add_cli_args(parser)
    parser = AsyncEngineArgs.add_cli_args(parser)
    args = parser.parse_args()
    engine_manager_args = EngineManagerArgs.from_cli_args(args)
    engine_args = AsyncEngineArgs.from_cli_args(args)

    check_engine_args(engine_args, engine_manager_args)

    print("engine_args: {}".format(engine_args))

    if args.launch_ray_cluster:
        # Launch the ray cluster for multi-node serving.
        launch_ray_cluster(args.ray_cluster_port)

    # Connect to a ray cluster.
    connect_to_ray_cluster(port=args.ray_cluster_port)

    # if gpu is not available, it means that this node is head pod without any llumnix components
    if is_gpu_available():
        # Launch the Llumnix componets on current node.
        server_id = random_uuid()
        node_id = ray.get_runtime_context().get_node_id()
        engine_manager, instance_ids, llumlets, request_output_queue = \
            init_llumnix_components(engine_manager_args, engine_args, node_id)

        for idx, ins_id in enumerate(instance_ids):
            instances[ins_id] = llumlets[idx]
            instance_num_requests[ins_id] = 0
        log_requests = not args.disable_log_requests_server
        # Start the api server after all the components of llumnix are ready.
        print(f"Start Api Server on '{args.host}:{args.port}'")
        uvicorn.run(app,
                    host=args.host,
                    port=args.port,
                    log_level="debug",
                    timeout_keep_alive=TIMEOUT_KEEP_ALIVE,
                    ssl_keyfile=args.ssl_keyfile,
                    ssl_certfile=args.ssl_certfile)
